// Copyright (c) 2014, The Shelf Bind project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf.bind.test;

import 'package:shelf_bind/shelf_bind.dart' as sb;
import 'package:shelf/shelf.dart' as s;
import 'package:shelf_route/shelf_route.dart' as sr;


import 'package:unittest/unittest.dart';
import 'package:mock/mock.dart';
import 'dart:convert';
import 'dart:async';

abstract class HandlerClass {
  dynamic call(s.Request request) { return null; }
}

class MockHandler extends Mock implements HandlerClass {
  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}

main() {
  MockHandler mockHandler;
  setUp(() {
    mockHandler = new MockHandler();
  });

  group('bind', () {

    void runTests(Type personType, sb.BindMode bindMode) {
      s.Handler _bindHandler() => sb.bind(personType, {'name': #name},
          mockHandler, bindMode: bindMode);

      test('should bind to variable in uri path', () {
        final Uri uri = Uri.parse('http://localhost/person/fred');
        final bindHandler = _bindHandler();
        final routeHandler = (sr.router()..get('/person/{name}', bindHandler))
            .handler;

        final request = new s.Request('GET', uri);
        routeHandler(request);

        mockHandler.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });

      test('should bind to variable in query parameter', (){
        final Uri uri = Uri.parse('http://localhost/person?name=fred');
        final bindHandler = _bindHandler();
        final routeHandler = (sr.router()..get('/person{?name}', bindHandler))
            .handler;

        final request = new s.Request('GET', uri);
        routeHandler(request);

        mockHandler.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });

    }

    group('in constructor mode', () {
      runTests(ConstructorPerson, sb.BindMode.CONSTRUCTOR);
    });

    group('in property mode with variable', () {
      runTests(FieldPerson, sb.BindMode.PROPERTY);
    });

    group('in property mode with setter', () {
      runTests(SetterPerson, sb.BindMode.PROPERTY);
    });

  });

  group('bindJsonBody', () {
    test('should bind to body', () {
      final Uri uri = Uri.parse('http://localhost/person');
      final bindHandler = sb.bindJsonBody(ConstructorPerson, mockHandler);
      final routeHandler = (sr.router()..post('/person', bindHandler))
          .handler;

      final json = JSON.encode(new ConstructorPerson('fred').toJson());
      final body = new Future.sync(() => json.codeUnits).asStream();
      final request = new s.Request('POST', uri, body: body);
      final responseFuture = new Future.sync(() => routeHandler(request))
      .then((_) {
        mockHandler.calls('call', personWithName(equals('fred')))
            .verify(happenedOnce);
      });
      expect(responseFuture, completes);
    });
  });
}

s.Response _handlePerson(AbstractPerson person, s.Request request) {
  print(person);
  return new s.Response.ok('yippee - ${person.toJson()}');
}

Matcher personWithName(matcher) => _personMatcher("name", matcher,
    (AbstractPerson p) => p.name);

Matcher _personMatcher(String fieldName, matcher, Getter getter)
    => fieldMatcher("Person", fieldName, matcher, getter);

Matcher fieldMatcher(String className, String fieldName, matcher,
                     Getter getter) =>
    new FieldMatcher(className, fieldName, matcher, getter);


typedef Getter(object);

class FieldMatcher extends CustomMatcher {
  final Getter getter;

  FieldMatcher(String className, String fieldName, matcher, this.getter)
      : super("$className with $fieldName that", fieldName, matcher);

  featureValueOf(actual) => getter(actual);
}


abstract class AbstractPerson {
  String get name;

  Map toJson() => { 'name': name };

  String toString() => '$runtimeType[name: $name]';
}

class ConstructorPerson extends AbstractPerson {
  final String name;

  ConstructorPerson(this.name);

  ConstructorPerson.build({String name}) : this(name);

  ConstructorPerson.fromJson(Map json) : this(json['name']);
}

class FieldPerson extends AbstractPerson {
  String name;

  FieldPerson();

  FieldPerson.fromJson(Map json) : this.name = json['name'];
}

class SetterPerson extends AbstractPerson {
  String _name;

  void set name(String name) {
    _name = name;
  }

  String get name => _name;

  SetterPerson();

  SetterPerson.fromJson(Map json) { this.name = json['name']; }
}